#include "stub.h"

ExecutionQueue::ExecutionQueue()
{
  for(int i = 0; i < MAX_EXECQUEUE; i++)
    {
      queue[i].active = false;
      pthread_mutex_init(&queue[i].joblock, NULL);
    }

  pthread_mutex_init(&broadcast_lock, NULL);
  pthread_cond_init(&broadcast, NULL);
}

ExecutionQueue::~ExecutionQueue()
{
  //free any unclaimed data (incase of aborting job)
  for(int i = 0; i < MAX_EXECQUEUE; i++)
    {
      if(queue[i].active && queue[i].data_length > 0)
	delete [] queue[i].data;
      pthread_mutex_destroy(&queue[i].joblock);
    }

  pthread_mutex_destroy(&broadcast_lock);
  pthread_cond_destroy(&broadcast);
}

ExecQueueMember *ExecutionQueue::getJob(int jobnum)
{
  for(;;)
    {
      //if queue is empty, keep checking
      pthread_mutex_lock(&queue[jobnum].joblock);
      if(queue[jobnum].active)
	{
	  //don't unlock -will unlock when done
	  ExecQueueMember *retval;
	  retval = &queue[jobnum];
	  return retval;
	}
      pthread_mutex_unlock(&queue[jobnum].joblock);

      //wait for next block of data to come in
      pthread_cond_wait(&broadcast, &broadcast_lock);
      pthread_mutex_unlock(&broadcast_lock);

    }
  return NULL;
}

void ExecutionQueue::markJobComplete(int jobnum)
{
  queue[jobnum].active = false;
  pthread_mutex_unlock(&queue[jobnum].joblock);
}

void ExecutionQueue::addNewJob(ExecQueueMember *j)
{
  //get next avaliable thread
  int i;
  //keep checking incase it doesn't have any
  for(;;)
    {
      for(i = 0; i < MAX_EXECQUEUE; i++)
	{
	  //see if in use
	  if(queue[i].active
	     || pthread_mutex_trylock(&queue[i].joblock) != 0)
	    continue;
	
	  queue[i].active = true;
	  queue[i].connection = j->connection;
	  queue[i].func_enum = j->func_enum;
	  queue[i].wait_id = j->wait_id;
	  queue[i].data_length = j->data_length;
	  queue[i].data = j->data;
	  pthread_mutex_unlock(&queue[i].joblock);

	  //announce to all that a new resource is here
	  pthread_cond_broadcast(&broadcast);

	  return;
	}
    }
}

long ExecutionQueue::getQueueLength()
{
  int numactive = 0;
  for(int i = 0; i < MAX_EXECQUEUE; i++)
    if(queue[i].active)
      numactive++;
  return numactive;
}


WaitForQueue::WaitForQueue()
{
  for(int i = 0; i < MAX_WAITQUEUE; i++)
    queue[i].active = false;
  pthread_mutex_init(&queuelock, NULL);
}

WaitForQueue::~WaitForQueue()
{
  pthread_mutex_destroy(&queuelock);
}

bool WaitForQueue::inQueue(long wait_id)
{
  pthread_mutex_lock(&queuelock);
  for(int i = 0; i < MAX_WAITQUEUE; i++)
    {
      if(queue[i].active && queue[i].wait_id == wait_id)
	{
	  pthread_mutex_unlock(&queuelock);
	  return true;
	}
    }

  pthread_mutex_unlock(&queuelock);
  return false;
}

long WaitForQueue::getReturnData(long wait_id)
{
  pthread_mutex_lock(&queuelock);

  for(int i = 0; i < MAX_WAITQUEUE; i++)
    {
      if(queue[i].active && queue[i].wait_id == wait_id)
	{
	  //must get values before unlocking just incase
	  long retval = queue[i].return_data;
	  queue[i].active = false;
	  pthread_mutex_unlock(&queuelock);
	  return retval;
	}
    }

  pthread_mutex_unlock(&queuelock);
  return 0; //something wrong happened...
}

bool WaitForQueue::StoreReturnValue(long wait_id, long return_data)
{
  //add new data
  int i;
  do //want to loop if list is full
    {

      pthread_mutex_lock(&queuelock);
      for(i = 0; i < MAX_WAITQUEUE; i++)
	{
	  if(queue[i].active)
	    continue;
	  queue[i].wait_id = wait_id;
	  queue[i].return_data = return_data;
	  queue[i].active = true;
	  break;
	}
      pthread_mutex_unlock(&queuelock);

    } while(i == MAX_WAITQUEUE);

  printf("recieved a return: wait_id: %d, value %d\n", wait_id, return_data);
  return true;
}


PeerConnections::PeerConnections(int instance)
{
  num_active_inconn = num_active_outconn = num_conn = 0;

  //load connection list from file
  FILE *f;
  f = fopen("connections.dat", "rb");

  char buffer[256];

  //instance reading data from file
  int current_instance = -1;

  while(!feof(f))
    {
      //read in line
      int i;
      buffer[0] = '\0'; //incase none read
      for(i = 0; i < 256; i++)
	{
	  fread(&buffer[i], 1, 1, f);
	  if(buffer[i] == '\n' || buffer[i] == '\0')
	    break;
	}
      buffer[i] = '\0';

      //ignore comments
      if( (buffer[0] == '/' && buffer[1] == '/')
	  || buffer[0] == '\0')
	continue;

      //see if instance block -if so, update
      if(buffer[0] == '#')
	{
	  current_instance = atoi(&buffer[1]);
	  continue;
	}

      //ignore other instances' information
      if(current_instance != instance)
	continue;

      //this is a piece of data we want
      //create a new server based on it
      createNewServer(buffer);
    }

  fclose(f);
}

PeerConnections::~PeerConnections()
{
  //close all connections
  for(int i = 0; i < num_conn * 2; i++)
    close(connection[i]);

  //terminate threads?
}

void PeerConnections::waitForConnections()
{
  //ahh, the joys of multi-threading :-)
  //  while(num_conn != num_activeconn);

  //**this doesn't need to be printed out -only for debug
  int i = 0;
  do {
    /*    printf("%d: %d,%d /  %d connections\n", i++,
	   num_active_inconn, num_active_outconn, num_conn);
	   sleep(1);*/
  } while(num_conn != num_active_inconn || num_conn != num_active_outconn);
  
  /*  printf("got all %d: %d,%d /  %d connections\n", i++,
      num_active_inconn, num_active_outconn, num_conn);*/
    printf("recieved all connections\n");
}

//creates connections to all peers (for sending)
void PeerConnections::createPeerConnections()
{
  for(int i = 0; i < num_conn; i++)
    {
      printf("connecting to: %s\n", conn_descriptor[i]);
      connection[(i << 1)+1] = CreateClientConnection(conn_descriptor[i]);
      printf("connected\n");
      num_active_outconn++;
    }
}

void PeerConnections::createNewServer(char *descriptor)
{
  //keep a local copy of this connection, and increment total
  long connection_num = num_conn++;

  conn_descriptor[connection_num] = new char[strlen(descriptor)+1];
  strcpy(conn_descriptor[connection_num], descriptor);
  
  int serverSocket = 0,
    on = 0,
    port = 0,
    status = 0,
    childPid = 0;
  struct hostent *hostPtr = NULL;
  char hostname[80] = "";
  struct sockaddr_in serverName = { 0 };

  //descriptor is in the format:
  //local_recieving_port:sending_ip:sending_port
  //grab port number
  port = atoi(descriptor);

  serverSocket = socket(PF_INET, SOCK_STREAM, 
			IPPROTO_TCP);

  if(-1 == serverSocket)
    {
      perror("socket()");
      //exit(1);
    }

  /* turn off bind address checking and allow 
     port numbers to be reused - otherwise
     the TIME_WAIT phenomenon will prevent 
     binding to these address.port combinations
     for (2 * MSL) seconds.  */

  on = 1;

  status = setsockopt(serverSocket, SOL_SOCKET, 
		      SO_REUSEADDR,
		      (const char *) &on, sizeof(on));
  if(-1 == status)
    {
      perror("setsockopt(...,SO_REUSEADDR,...)");
    }

  /* when connection is closed, need 
     to linger to ensure all data is transmitted */
  {
    struct linger linger = { 0 };
    
    linger.l_onoff = 1;
    linger.l_linger = 30;
    status = setsockopt(serverSocket, 
			SOL_SOCKET, SO_LINGER,
			(const char *) &linger, 
			sizeof(linger));    
    if(-1 == status)
      {
	perror("setsockopt(...,SO_LINGER,...)");
      }
  }

  //get self
  status = _GetHostName(hostname, 
			sizeof(hostname));
  if(-1 == status)
    {
      perror("_GetHostName()");
      //exit(1);
    }

  hostPtr = gethostbyname(hostname);
  if(NULL == hostPtr)
    {
      perror("gethostbyname()");
      //exit(1);
    }

  (void)memset(&serverName, 0, 
		sizeof(serverName));
  (void)memcpy(&serverName.sin_addr, 
		hostPtr->h_addr,
		hostPtr->h_length);

  // to allow server be contactable on any of 
  //   its IP addresses, uncomment the following 
  //  line of code:
  //   serverName.sin_addr.s_addr=htonl(INADDR_ANY);
   

  serverName.sin_family = AF_INET;
  //network order
  serverName.sin_port = htons(port);

  status = bind(serverSocket, 
		(struct sockaddr *) &serverName,
		sizeof(serverName));
  if(-1 == status)
    {
      perror("bind()");
      //exit(1);
    }

  status = listen(serverSocket, BACK_LOG);
  if(-1 == status)
    {
      perror("listen()");
      //exit(1);
    }

  //create thread to wait for other end's connection and put connection in
  //correct spot of array (use connection_num for where to put in array)
  //put serverSocket in array, will be closed by server thread
  //and will be replaced by the slaveSocket -the one to be written to
  connection[(connection_num << 1)] = serverSocket;
  printf("creating thread to monitor port %d\n", port);
  pthread_create(&clientmonitor[connection_num], NULL, AcceptServerConnection,
  		 &connection[connection_num << 1]);
}

int PeerConnections::matchingSendSocket(int socket)
{
  for(int i = 0; i < num_conn; i++)
    {
      if(connection[i << 1] == socket)
	return connection[(i << 1) + 1];
    }
  return 0; //error
}

int PeerConnections::sendSocketWithShortQueue()
{
  long minqueue = MAX_EXECQUEUE;
  long minconn = 0; //connection with minimum queue
  char buffer[8];
  int bufferpos = 0;
  int iteration = 0; //used for keeping track of waiting
  //must find a server to send job to
  //if not, wait for one
  do
    {
      for(int i = 0; i < num_conn; i++)
	{
	  write(connection[(i << 1)+1], "q", 1);
	  printf("queue query sent to connection %d\n", i);
	  do
	    {
	      bufferpos += read(connection[(i << 1) + 1], &buffer[bufferpos], 5 - bufferpos);
	    } while(bufferpos < 5);
	  if(buffer[0] != 'p')
	    printf("invalid response to queue query\n");
	  long queuelen = *((long *)(&buffer[1]));
	  if(queuelen < minqueue)
	    {
	      minqueue = queuelen;
	      minconn = i;
	    }
	}
      if(iteration != 0)
	{
	  //send message that this happened
	  putstring("WARNING: node is deadended (increase number of threads)\n");
	  sleep(1);
	}
      iteration++;
    } while(minqueue >= MAX_EXECQUEUE);

  return connection[minconn+1];
}
