#include "common.h"


Optimizer::Optimizer(SymbolTable *yt)
{
	symboltable = yt;
}


//combines integer constants during compile time
//so they do not need to be done during runtime
void Optimizer::ConstantFolding(ASTNode *&tree)
{
	if(tree == NULL)
		return;

	//don't want to do anything with variables...
	if((signed long)tree->op >= (signed long)T_IDENTIFIERSTART)
		return;


	if(tree->child[0] != NULL)
		ConstantFolding(tree->child[0]);
	if(tree->child[1] != NULL)
		ConstantFolding(tree->child[1]);
	if(tree->child[2] != NULL)
		ConstantFolding(tree->child[2]);

	//don't want anything like this
	if(tree->child[0] == NULL)
		return;

	//this is useless, so get rid of it
	if(tree->op == T_UNARYPLUS)
	{
		ASTNode *temp = tree;
		tree = tree->child[0];
		tree->parent = temp->parent;
		delete temp;
		return;
	}

	//if the first isn't a constant, then we can exit early
	if(tree->child[0]->op > T_TYPE_MAX)
		return;

	if(tree->op == T_UNARYMINUS)
	{
		ASTNode *temp = tree;
		tree = tree->child[0];
		tree->parent = temp->parent;
		delete temp;
		//take negative
		switch(tree->op & A_BASETYPES)
		{
		case T_DOUBLE:
			*(double *)tree->data = -(*(double *)tree->data);
			break;
		case T_FLOAT:
			*(float *)tree->data = -(*(float *)tree->data);
			break;
		default:
			*(long *)tree->data = -(*(long *)tree->data);
		}
		return;
	}

	if(tree->op == T_BNOT)
	{
		ASTNode *temp = tree;
		tree = tree->child[0];
		tree->parent = temp->parent;
		delete temp;
		//take logical 'not'
		//assume long, because it really should be!
		*(long *)tree->data = ~(*(long *)tree->data);
		return;
	}

	//now check to see if 2nd constant
	if(tree->child[1] == NULL || tree->child[1]->op > T_TYPE_MAX)
		return;

	//make sure both are integers
	if(tree->child[0]->op == T_FLOAT || tree->child[0]->op == T_DOUBLE
			|| tree->child[1]->op == T_FLOAT || tree->child[1]->op == T_DOUBLE)
		return;

	//fold constant
	ASTNode *temp = tree;
	tree = tree->child[0];
	tree->parent = temp->parent;

	switch(temp->op)
	{
	case T_PLUS:
		*(long *)tree->data += (*(long *)temp->child[1]->data);
		break;
	case T_MINUS:
		*(long *)tree->data -= (*(long *)temp->child[1]->data);
		break;
	case T_MULTIPLY:
		*(long *)tree->data *= (*(long *)temp->child[1]->data);
		break;
	case T_DIVIDE:
		//if divide by zero, MUST do it in realtime to generate error
		if(*(long *)temp->child[1]->data == 0)
		{
			tree->parent = temp;
			tree = temp;
			return;
		}
		*(long *)tree->data /= (*(long *)temp->child[1]->data);
		break;
	case T_MODULUS:
		//if divide by zero, MUST do it in realtime to generate error
		if(*(long *)temp->child[1]->data == 0)
		{
			tree->parent = temp;
			tree = temp;
			return;
		}
		*(long *)tree->data %= (*(long *)temp->child[1]->data);
		break;
	case T_SHIFTLEFT:
		*(long *)tree->data <<= (*(long *)temp->child[1]->data);
		break;
	case T_SHIFTRIGHT:
		*(long *)tree->data >>= (*(long *)temp->child[1]->data);
		break;

	case T_BOR:
		*(long *)tree->data |= (*(long *)temp->child[1]->data);
		break;
	case T_BAND:
		*(long *)tree->data &= (*(long *)temp->child[1]->data);
		break;

	case T_EQUALS:
		*(long *)tree->data = (*(long *)tree->data == (*(long *)temp->child[1]->data));
		break;
	case T_LESS:
		*(long *)tree->data = (*(long *)tree->data < (*(long *)temp->child[1]->data));
		break;
	case T_LEQUAL:
		*(long *)tree->data = (*(long *)tree->data <= (*(long *)temp->child[1]->data));
		break;
	case T_GREATER:
		*(long *)tree->data = (*(long *)tree->data > (*(long *)temp->child[1]->data));
		break;
	case T_GEQUAL:
		*(long *)tree->data = (*(long *)tree->data >= (*(long *)temp->child[1]->data));
		break;
	case T_NOTEQUAL:
		*(long *)tree->data = (*(long *)tree->data != (*(long *)temp->child[1]->data));
		break;

	default:
		break;
	}

	delete temp->child[1];
	delete temp;
	return;
}


//combines integer constants during compile time
//so they do not need to be done during runtime
//including addresses
void Optimizer::ConstantFolding(MIRInstr *mir)
{
	//return;

	//check each instruction
	for(MIRInstr *m = mir; m != NULL; m = m->next)
	{
		if(! ((m->op1_id & T_IMMEDIATE) && (m->op2_id & T_IMMEDIATE)) )
			continue;

		//don't want to optimize non-integers
		if(m->op1_id == (T_IMMEDIATE | T_FLOAT) || m->op1_id == (T_IMMEDIATE | T_DOUBLE)
			|| m->op2_id == (T_IMMEDIATE | T_FLOAT) || m->op2_id == (T_IMMEDIATE | T_DOUBLE)
			|| m->dest_pointer)
			continue;

		bool keep_original_statement = false;	//only used if need for pointer

		switch(m->instr)
		{
		case T_PLUS:
			*(long *)m->op1_data += (*(long *)m->op2_data);
			break;
		case T_MINUS:
			*(long *)m->op1_data -= (*(long *)m->op2_data);
			break;
		case T_MULTIPLY:
			*(long *)m->op1_data *= (*(long *)m->op2_data);
			break;
		case T_DIVIDE:
			//if divide by zero, MUST do it in realtime to generate error
			if(*(long *)m->op2_data == 0)
				continue;
			*(long *)m->op1_data /= (*(long *)m->op2_data);
			break;
		case T_MODULUS:
			//if divide by zero, MUST do it in realtime to generate error
			if(*(long *)m->op2_data == 0)
				continue;
			*(long *)m->op1_data %= (*(long *)m->op2_data);
			break;
		case T_SHIFTLEFT:
			*(long *)m->op1_data <<= (*(long *)m->op2_data);
			break;
		case T_SHIFTRIGHT:
			*(long *)m->op1_data >>= (*(long *)m->op2_data);
			break;

		case T_BOR:
			*(long *)m->op1_data |= (*(long *)m->op2_data);
			break;
		case T_BAND:
			*(long *)m->op1_data &= (*(long *)m->op2_data);
			break;

		case T_EQUALS:
			*(long *)m->op1_data = (*(long *)m->op1_data == (*(long *)m->op2_data));
			break;
		case T_LESS:
			*(long *)m->op1_data = (*(long *)m->op1_data < (*(long *)m->op2_data));
			break;
		case T_LEQUAL:
			*(long *)m->op1_data = (*(long *)m->op1_data <= (*(long *)m->op2_data));
			break;
		case T_GREATER:
			*(long *)m->op1_data = (*(long *)m->op1_data > (*(long *)m->op2_data));
			break;
		case T_GEQUAL:
			*(long *)m->op1_data = (*(long *)m->op1_data >= (*(long *)m->op2_data));
			break;
		case T_NOTEQUAL:
			*(long *)m->op1_data = (*(long *)m->op1_data != (*(long *)m->op2_data));
			break;

		default:
			continue;
		}

		//keep this just incase we keep the instruction
		m->instr = T_PLUS;
		*(long *)m->op2_data = 0;

		if(m->dest_id == T_TEMPREG && !m->dest_pointer)
		{
			//replace all of this temporary until it's assigned to
			unsigned long replacedefnum = m->dest_defnum;
			for(MIRInstr *cur = m->next; cur != NULL; cur = cur->next)
			{
				//check to see if assigned to, if so, then the value is changed and we must stop
				if(cur->dest_id == T_TEMPREG && cur->dest_defnum == replacedefnum)
				{
					if(cur->dest_pointer)
					{
						keep_original_statement = true;
					}
					else
						break;
				}

				if(cur->op1_id == T_TEMPREG && cur->op1_defnum == replacedefnum)
				{
					cur->op1_id = (T_IMMEDIATE | T_LONG);
					cur->op1_data = (void *)new long;
					*(long *)cur->op1_data = *(long *)m->op1_data;
				}

				if(cur->op2_id == T_TEMPREG && cur->op2_defnum == replacedefnum)
				{
					cur->op2_id = (T_IMMEDIATE | T_LONG);
					cur->op2_data = (void *)new long;
					*(long *)cur->op2_data = *(long *)m->op1_data;
				}
			}
			//delete m->op1_data;
			//delete m->op2_data;
			MIRInstr *temp = m;
			if(m->prev == NULL) //check for this odd case
			{
				mir = temp->next;
				m = mir;
				continue;
			}

			if(!keep_original_statement)
			{
				m = m->prev;
				m->next = temp->next;
				temp->next->prev = m;
				delete temp;
			}
		}

	}

	return;
}


//removes goto's proceding goto's (without a label)
//should be done near the very end (after loop optimizers)
void Optimizer::RemoveRedundantGOTOs(MIRInstr *mir)
{
	if(mir == NULL)
		return;

	while(mir->next != NULL)
	{
		if(mir->instr == T_GOTO && mir->next->instr == T_GOTO)
		{
			MIRInstr *temp = mir->next;
			mir->next = temp->next;
			mir->next->prev = mir;
			delete temp;
			continue;
		}

		mir = mir->next;
	}
}

