#include "common.h"

////////////////
//OpStack class

OpStack::OpStack()
{
	root = NULL;
}


OpStack::~OpStack()
{
	//only enters while loop when errors in file
	while(root != NULL)
	{
		OpDef *temp = root->next;
		delete root;
		root = temp;
	}
}

void OpStack::pushOp(ASTNode *n, long parendepth, long bracketdepth)
{
	OpDef *temp = new OpDef;
	temp->astnode = n;
	temp->parendepth = parendepth;
	temp->bracketdepth = bracketdepth;
	temp->next = root;
	root = temp;
}

void OpStack::pushOp(ASTNode *n)
{
	OpDef *temp = new OpDef;
	temp->astnode = n;
	temp->parendepth = 0;
	temp->bracketdepth = 0;
	temp->next = root;
	root = temp;
}

ASTNode *OpStack::popOp()
{
	if(root == NULL)
		return NULL;

	ASTNode *astnode = root->astnode;
	OpDef *temp = root;
	root = root->next;
	delete temp;
	return astnode;
}

ASTNode *OpStack::peekOp()
{
	if(root == NULL)
		return NULL;
	return root->astnode;

}

long OpStack::peekParenDepth()
{
	return root->parendepth;
}
long OpStack::peekBracketDepth()
{
	return root->bracketdepth;
}

////////////////////
//VarDefStack class

VarDefStack::VarDefStack()
{
	root = NULL;
}

VarDefStack::~VarDefStack()
{
	while(root != NULL)
	{
		VarDef *temp = root->next;
		delete root;
		root = temp;
	}
}

void VarDefStack::pushVarDef(TOKEN_ID id, long curdefnum)
{
	VarDef *temp = new VarDef;
	temp->id = id;
	temp->curdefnum = curdefnum;
	temp->next = root;
	root = temp;
}

void VarDefStack::increaseScope()
{
	VarDef *temp = new VarDef;
	temp->id = SCOPECHANGE;
	temp->curdefnum = SCOPECHANGE;
	temp->next = root;
	root = temp;
}

//flushes variables when exiting scope
void VarDefStack::decreaseScope()
{
	//delete all prior to scope change
	while(root != NULL && root->id != SCOPECHANGE)
	{
		VarDef *temp = root->next;
		delete root;
		root = temp;
	}
	//now delete scope change
	if(root != NULL)
	{
		VarDef *temp = root->next;
		delete root;
		root = temp;
	}
}

//returns first defnum
long VarDefStack::findFirstDef(TOKEN_ID id)
{
	VarDef *temp = root;
	while(temp != NULL)
	{
		if(temp->id == id)
		{
			return temp->curdefnum;
		}
		temp = temp->next;
	}

	return 0;
}

//returns defnum in top scope
bool VarDefStack::ExistInCurScope(TOKEN_ID id)
{
	VarDef *temp = root;
	while(temp != NULL && temp->id != SCOPECHANGE)
	{
		if(temp->id == id)
		{
			return true;
		}
		temp = temp->next;
	}

	return false;
}



/////////////
//parser class

Parser::Parser(char *file, StringTable *st, SymbolTable *yt)
{
	stringtable = st;
	symboltable = yt;
	scanner = new Scanner(file, stringtable);
	constrainer = new Constrainer(scanner);

	curScopeDepth = 0;	//set to global
	opstack = new OpStack;
	varstack = new OpStack;
	vardefstack = new VarDefStack;

	globalvarsize = 0;
}

Parser::~Parser()
{
	delete scanner;
	delete opstack;
	delete varstack;
	delete vardefstack;
};


bool Parser::EndParse()
{
	constrainer->finalReport();
	return constrainer->validForCompile();
}

///////////////////////////////
//Parse

bool Parser::Parse()
{
	curattrib = A_SIGNED;
	curparent = GLOBALSCOPE;	//begin global
	curparentdefnum = 0;	//to know where to add to class
	inclass = false;	//begin global
	require_func_definition = false;	//only for scoped function definitions

	//loop for declarations (different loop for within functions)
	for(;;)
	{
		curtoken = scanner->getNextToken();

		//first character from global perspective must be type (variable/function declaration)
		switch(curtoken)
		{

		//******
		//final type definitions
		case T_CHAR:
		case T_SHORT:
		case T_LONG:
		case T_FLOAT:
		case T_DOUBLE:
		case T_VOID:
			if(!ParseGlobalTypeDefinition())
				return EndParse();
			break;

		//*******
		//classes
		case T_CLASS:
		case T_PCLASS:
			if(!ParseGlobalClassDefinition())
				return EndParse();
			break;

		//******
		//end class
		case T_RBRACE:
			if(inclass)
			{
				//exit class back to global
				curparent = GLOBALSCOPE;
				vardefstack->decreaseScope();	//remove class members from globals
				inclass = false;
			}
			else
				constrainer->reportError(ERROR_INVALIDSYMBOL);
			break;

		//*******
		//qualifiers
		case T_CONST:
			if(curattrib == A_SIGNED)
				curattrib = A_CONST;
			else	//const must be first part of declaration
				constrainer->reportError(ERROR_REPEATEDQUALIFIER);
			break;

		case T_EXTERN:
			if(curattrib == A_SIGNED)
				curattrib = A_EXTERNAL;
			else	//const must be first part of declaration
				constrainer->reportError(ERROR_REPEATEDQUALIFIER);
			break;


		case T_UNSIGNED:
			if((curattrib & A_UNSIGNED) != A_UNSIGNED)
				curattrib |= A_UNSIGNED;
			else	//const must be first part of declaration
				constrainer->reportError(ERROR_REPEATEDQUALIFIER);
			break;


		//*****
		//end of file
		case T_ENDOFFILE:
			//put code here to finish constructing ast
			return EndParse();

		//*****
		//errors
		case T_RECURSEINCLUDE:
			constrainer->reportError(ERROR_RECURSIVEINCLUDE);
			break;

		//*****
		//ignore
		case T_SEMICOLON:	//ignore according to language
			break;

		//*****
		//user types or errors
		default:
			if(!ParseGlobalUserTypeDefinition())
				return EndParse();
			break;	//default

		} //switch

	}//infinite loop

	return false;
}

//////////////////////
//parse definition of variables, arrays, functions, etc.
bool Parser::ParseGlobalTypeDefinition()
{
	//save type
	curattrib |= curtoken;

	//get name
	curtoken = scanner->getNextToken();

	//make sure not error
	if(curtoken == T_ENDOFFILE)
	{
		constrainer->reportError(ERROR_UNEXPECTEDENDOFFILE);
		return false;
	}
	else if(curtoken == T_INVALID || curtoken < T_BUILTINFUNCSTART)
	{
		constrainer->reportError(ERROR_INVALIDIDENTIFIER);
		return true;	//continue on parsing
	}


	symbol_already_exists = vardefstack->ExistInCurScope(curtoken);

	switch(scanner->getNextToken())
	{
	//************
	//variable declaration
	case T_SEMICOLON:
		if(!ParseGlobalVarDefinition())
			return false;
		break;

	//***********
	//initialized variable
	case T_GETS:
		if(curattrib & A_EXTERNAL)
			constrainer->reportError(ERROR_INITIALIZEDEXTERNAL);
		if(!ParseGlobalVarDefinitionInit())
			return false;
		break;

	//**********
	//array
	case T_LBRACKET:
		if(!ParseGlobalArrayDefinition())
			return false;
		break; //array

	//********
	//function definition
	case T_LPAREN:
		if(!ParseGlobalFunction())
			return false;
		break;

	//********
	//member function declaration
	case T_SCOPE_RES:
		if(curattrib & A_EXTERNAL)
			constrainer->reportError(ERROR_SCOPEDEXTERNAL);
		if(!ParseGlobalClassScopeFunction())
			return false;
		break;


	case T_ENDOFFILE:
		constrainer->reportError(ERROR_UNEXPECTEDENDOFFILE);
		return false;

	default:
		constrainer->reportError(ERROR_INVALIDSYMBOL);
		break;
	}

	//return to default
	curattrib = A_SIGNED;
	return true;
}

//////////////////////
//parse global variable definitions
bool Parser::ParseGlobalVarDefinition()
{
	if(symbol_already_exists)
		constrainer->reportError(ERROR_SYMBOLDEFINED);

	//make sure not a constant (because constants should be initialized
	if(curattrib & A_CONST)
		constrainer->reportError(ERROR_CONSTANTNOTINIT);

	long defnum = symboltable->getNextDefNum(curtoken);
	vardefstack->pushVarDef(curtoken, defnum);
	symboltable->addSymbol(curtoken, defnum, scanner->getEndNameofLastGet(),
		curattrib, 0, 0, (unsigned char *)"\0\0\0\0\0\0\0\0", curparent);

	if(inclass)
	{
		//keep track of position within class
		symboltable->setAddress(curtoken, defnum, symboltable->getSize(curparent, curparentdefnum));
		//add size to class
		symboltable->addSize(curparent, curparentdefnum, symboltable->getSize(curtoken, defnum));
	}
	else //global
	{
		//keep track of position relative to globals
		symboltable->setAddress(curtoken, defnum, globalvarsize);
		//add size to global size
		globalvarsize += symboltable->getSize(curtoken, defnum);
	}

	return true;
}

//////////////////////
//parse global variable definitions with initialization
bool Parser::ParseGlobalVarDefinitionInit()
{
	if(inclass)
		constrainer->reportError(ERROR_MEMBERINITIALIZATION);
	if(symbol_already_exists)
	{
		constrainer->reportError(ERROR_SYMBOLDEFINED);
	}

	long defnum = symboltable->getNextDefNum(curtoken);
	vardefstack->pushVarDef(curtoken, defnum);
	symboltable->addSymbol(curtoken, defnum, scanner->getEndNameofLastGet(),
		curattrib, 0, 0, (unsigned char *)"\0\0\0\0\0\0\0\0", curparent);

	//don't really need to check if inclass, but do anyway to keep code consistent
	if(inclass)
	{
		//keep track of position within class
		symboltable->setAddress(curtoken, defnum, symboltable->getSize(curparent, curparentdefnum));
		//add size to class
		symboltable->addSize(curparent, curparentdefnum, symboltable->getSize(curtoken, defnum));
	}
	else //global
	{
		//keep track of position relative to globals
		symboltable->setAddress(curtoken, defnum, globalvarsize);
		//add size to global size
		globalvarsize += symboltable->getSize(curtoken, defnum);
	}

	nexttoken = scanner->getNextToken();

	int immediatesign = 1;
	//check to see if negative
	if(nexttoken == T_MINUS)
	{
		immediatesign = -1;
		nexttoken = scanner->getNextToken();
	}

	//make sure immediate
	if(nexttoken & T_IMMEDIATE)
	{
		//remove T_IMMEDIATE flag
		nexttoken &= (~(TOKEN_ID)T_IMMEDIATE);
		//if value is of different type, typecast
		unsigned char buffer[8];
		memset(buffer, 0, 8);
		switch(nexttoken)
		{
		case T_CHAR:
			buffer[0] = (char)scanner->getImmediateLong();
			buffer[0] *= immediatesign;
			symboltable->setInitVal(curtoken, defnum, buffer, T_CHAR);
			break;
		case T_SHORT:
			((short *)buffer)[0] = (short)scanner->getImmediateLong();
			((short *)buffer)[0] *= immediatesign;
			symboltable->setInitVal(curtoken, defnum, buffer, T_SHORT);
			break;
		case T_LONG:
			((long *)buffer)[0] = (long)scanner->getImmediateLong();
			((long *)buffer)[0] *= immediatesign;
			symboltable->setInitVal(curtoken, defnum, buffer, T_LONG);
			break;
		case T_FLOAT:
			((float *)buffer)[0] = (float)scanner->getImmediateDouble();
			((float *)buffer)[0] *= immediatesign;
			symboltable->setInitVal(curtoken, defnum, buffer, T_FLOAT);
			break;
		case T_DOUBLE:
			((double *)buffer)[0] = (double)scanner->getImmediateDouble();
			((double *)buffer)[0] *= immediatesign;
			symboltable->setInitVal(curtoken, defnum, buffer, T_DOUBLE);
			break;
		case T_IMMEDIATE_STR:
			constrainer->reportError(ERROR_INVALIDTYPE);
			break;
		case T_ENDOFFILE:
			constrainer->reportError(ERROR_UNEXPECTEDENDOFFILE);
			return false;
		default:
			constrainer->reportError(ERROR_INVALIDSYMBOL);
			break;
		}
	}
	else
		constrainer->reportError(ERROR_INITMUSTBECONSTANT);

	//get trailing semicolon
	nexttoken = scanner->getNextToken();
	//make sure not error
	if(curtoken == T_ENDOFFILE)
	{
		constrainer->reportError(ERROR_UNEXPECTEDENDOFFILE);
		return false;
	}
	if(nexttoken != T_SEMICOLON)
		constrainer->reportError(ERROR_INVALIDSYMBOL);

	return true;
}

/////////////////////
//parse global array definitions
bool Parser::ParseGlobalArrayDefinition()
{
	curattrib |= A_ARRAY;

	if(symbol_already_exists)
	{
		constrainer->reportError(ERROR_SYMBOLDEFINED);
	}

	long defnum = symboltable->getNextDefNum(curtoken);
	vardefstack->pushVarDef(curtoken, defnum);
	symboltable->addSymbol(curtoken, defnum, scanner->getEndNameofLastGet(),
		curattrib, 0, 0, (unsigned char *)"\0\0\0\0\0\0\0\0", curparent);

	nexttoken = scanner->getNextToken();
	bool userlength = false;
	long arraylength;
	//see if initialization specifies length
	if(nexttoken == T_RBRACKET)
	{
		//if external, this can't happen
		if(curattrib & A_EXTERNAL)
		{
			constrainer->reportError(ERROR_INITIALIZEDEXTERNAL);
			return false;
		}

		userlength = true;
		nexttoken = scanner->getNextToken();
		if(nexttoken != T_GETS)
			constrainer->reportError(ERROR_ARRAYNOLENGTHNOINIT);
	}
	else //length given by number
	{

		if(!(nexttoken & T_IMMEDIATE))
			constrainer->reportError(ERROR_CONSTANTREQUIRED);

		//remove immediate flag
		nexttoken &= (~(TOKEN_ID)T_IMMEDIATE);

		//check for type
		if(nexttoken != T_CHAR && nexttoken != T_SHORT && nexttoken != T_LONG)
			constrainer->reportError(ERROR_INVALIDTYPE);

		//get length of array
		arraylength = scanner->getImmediateLong();
		symboltable->setSize(curtoken, defnum, arraylength * symboltable->getSize(curtoken, defnum));
		if(arraylength == 0)
		{
			symboltable->setType(curtoken, defnum, curattrib | A_RESIZEABLE);
			symboltable->setSize(curtoken, defnum, SIZE_LONG);
		}

		//don't really need to check if inclass, but do anyway to keep code consistent
		if(inclass)
		{
			//keep track of position within class
			symboltable->setAddress(curtoken, defnum, symboltable->getSize(curparent, curparentdefnum));
			//add size to class
			symboltable->addSize(curparent, curparentdefnum, symboltable->getSize(curtoken, defnum));
		}
		else //global
		{
			//keep track of position relative to globals
			symboltable->setAddress(curtoken, defnum, globalvarsize);
			//add size to global size
			globalvarsize += symboltable->getSize(curtoken, defnum);
		}

		//get trailing bracket
		nexttoken = scanner->getNextToken();
		//make sure not error
		if(curtoken == T_ENDOFFILE)
		{
			constrainer->reportError(ERROR_UNEXPECTEDENDOFFILE);
			return false;
		}
		if(nexttoken != T_RBRACKET)
			constrainer->reportError(ERROR_BRACKETREQUIRED);

		nexttoken = scanner->getNextToken();
		//exit if semicolon
		if(nexttoken == T_SEMICOLON)
			return true;

		//can't be an initialized external
		if(curattrib & A_EXTERNAL)
		{
			constrainer->reportError(ERROR_INITIALIZEDEXTERNAL);
			return false;
		}

		//make sure not error
		if(curtoken == T_ENDOFFILE)
		{
			constrainer->reportError(ERROR_UNEXPECTEDENDOFFILE);
			return false;
		}

		if(nexttoken != T_GETS)
			constrainer->reportError(ERROR_INVALIDSYMBOL);
	}

	//now get init data
	nexttoken = scanner->getNextToken();
	long userarraylength;

	if(nexttoken == (T_IMMEDIATE | T_IMMEDIATE_STR))
	{
		if( (curattrib & A_BASETYPES) != T_CHAR)
		{
			constrainer->reportError(ERROR_INVALIDTYPE);
			return true;
		}
		char *strtemp = scanner->getImmediateString();
		symboltable->setInitVal(curtoken, defnum, (unsigned char *)strtemp, A_ARRAY | T_IMMEDIATE_STR);
		userarraylength = strlen(strtemp) + 1;
	}
	else if(nexttoken == T_LBRACE)
	{
		//get each member of initialization
		//assume no more than MAXARRAYINITSIZE entries (MAXARRAYINITSIZE * SIZE_MAXDATA bytes)
		unsigned char *buffer = new unsigned char[MAXARRAYINITSIZE*SIZE_MAXDATA];
		TOKEN_ID arraytype = (symboltable->getType(curtoken, defnum) & (~(TOKEN_ID)A_ARRAY));

		//get array list
		for(userarraylength = 0; ; userarraylength++)
		{
			if(userarraylength >= MAXARRAYINITSIZE)
			{
				constrainer->reportError(ERROR_ARYTOOLARGEFORCOMPILE);
				break;
			}

			nexttoken = scanner->getNextToken();
			//make sure not error
			if(nexttoken == T_ENDOFFILE)
			{
				constrainer->reportError(ERROR_UNEXPECTEDENDOFFILE);
				return false;
			}

			int immediatesign = 1;
			if(nexttoken == T_MINUS)
			{
				immediatesign = -1;
				nexttoken = scanner->getNextToken();
			}

			if(!(nexttoken & T_IMMEDIATE))
			{
				constrainer->reportError(ERROR_INVALIDSYMBOL);
			}
			//make sure correct type
			nexttoken &= (~(TOKEN_ID)T_IMMEDIATE);

			//select value to return
			if(arraytype == T_DOUBLE || arraytype == T_FLOAT)
			{
				double value;
				//typecast if different
				if(nexttoken == T_DOUBLE || nexttoken == T_FLOAT)
					value = immediatesign * scanner->getImmediateDouble();
				else //int
					value = immediatesign * (double)scanner->getImmediateLong();

				//set init to value
				if(arraytype == T_DOUBLE)
					((double *)buffer)[userarraylength] = value;
				else
					((float *)buffer)[userarraylength] = (float)value;
			}									
			else //int type
			{
				long value;
				//typecast if different
				if(nexttoken != T_DOUBLE && nexttoken != T_FLOAT)
					value = immediatesign * scanner->getImmediateLong();
				else
					value = immediatesign * (long)scanner->getImmediateDouble();

				//set init to value
				switch(arraytype)
				{
				case T_CHAR:
					((char *)buffer)[userarraylength] = (char)value;
					break;
				case T_SHORT:
					((short *)buffer)[userarraylength] = (short)value;
					break;
				case T_LONG:
					((long *)buffer)[userarraylength] = (long)value;
					break;
				default:
					((long *)buffer)[userarraylength] = (long)value;
					break;
				}
			} //else int type

			//get comma or brace
			nexttoken = scanner->getNextToken();
			//make sure not error
			if(nexttoken == T_ENDOFFILE)
			{
				constrainer->reportError(ERROR_UNEXPECTEDENDOFFILE);
				return false;
			}
			if(nexttoken == T_RBRACE)
			{
				userarraylength++;	//count current one
				break;	//all done
			}
			else if(nexttoken != T_COMMA)
			{
				constrainer->reportError(ERROR_INVALIDSYMBOL);
			}

		} //for array init list

		//have list
		symboltable->setInitVal(curtoken, defnum, buffer, A_ARRAY | T_IMMEDIATE_STR);
	}
	else
	{
		constrainer->reportError(ERROR_INVALIDSYMBOL);
	}

	////////
	//update length if userlength
	if(userlength)
	{
		symboltable->setSize(curtoken, defnum, userarraylength * symboltable->getSize(curtoken, defnum));
		
		if(inclass)
		{
			constrainer->reportError(ERROR_MEMBERINITIALIZATION);
		}
		else //global
		{
			//keep track of position relative to globals
			symboltable->setAddress(curtoken, defnum, globalvarsize);
			//add size to global size
			globalvarsize += symboltable->getSize(curtoken, defnum);
		}
	}
	//not userlength -check boudary
	else if(userarraylength > arraylength)
		constrainer->reportError(ERROR_ARRAYINITTOOLONG);

	//get trailing semicolon
	nexttoken = scanner->getNextToken();
	if(nexttoken != T_SEMICOLON)
		constrainer->reportError(ERROR_INVALIDSYMBOL);
	
	return true;
}

//////////////////////////
//start of class definitions
bool Parser::ParseGlobalClassDefinition()
{
	//make sure if class, not qualified
	if(curattrib != A_SIGNED)
		constrainer->reportError(ERROR_QUALIFIEDCLASS);

	if(inclass)
		constrainer->reportError(ERROR_NESTEDCLASSDEFINITION);

	//save type
	curattrib = curtoken;

	//get name
	curtoken = scanner->getNextToken();

	//make sure not error
	if(curtoken == T_ENDOFFILE)
	{
		constrainer->reportError(ERROR_UNEXPECTEDENDOFFILE);
		return false;
	}
	else if(curtoken == T_INVALID || curtoken < T_BUILTINFUNCSTART)
	{
		constrainer->reportError(ERROR_INVALIDIDENTIFIER);
		return true;	//continue on parsing
	}

	if(vardefstack->ExistInCurScope(curtoken))
		constrainer->reportError(ERROR_SYMBOLDEFINED);

	long defnum = symboltable->getNextDefNum(curtoken);
	vardefstack->pushVarDef(curtoken, defnum);
	symboltable->addSymbol(curtoken, defnum, scanner->getEndNameofLastGet(),
		curattrib, 0, 0, (unsigned char *)"\0\0\0\0\0\0\0\0", curparent);

	nexttoken = scanner->getNextToken();
	if(nexttoken != T_LBRACE)
		constrainer->reportError(ERROR_INVALIDSYMBOL);

	inclass = true;
	curparent = curtoken;	//so all future will be assigned to class, until exit class
	curparentdefnum = defnum;
	vardefstack->increaseScope();	//put all in class

	//return to default
	curattrib = A_SIGNED;

	return true;
}


//////////////////////////
//for user specified type declarations
bool Parser::ParseGlobalUserTypeDefinition()
{
	//check to see if instatiation of user type
	long curclassdefnum = vardefstack->findFirstDef(curtoken);
	if(curclassdefnum
		&& (symboltable->getType(curtoken, curclassdefnum) == T_CLASS || symboltable->getType(curtoken, curclassdefnum) == T_PCLASS) )
	{
		//instantiation

		//remember current class
		TOKEN_ID curclass = curtoken;
		curtoken = scanner->getNextToken();

		//make sure not error
		if(curtoken == T_ENDOFFILE)
		{
			constrainer->reportError(ERROR_UNEXPECTEDENDOFFILE);
			return false;
		}
		else if(curtoken == T_INVALID || curtoken < T_BUILTINFUNCSTART)
		{
			constrainer->reportError(ERROR_INVALIDIDENTIFIER);
			return true;	//continue on parsing
		}

		//check to make sure exists
		if(vardefstack->ExistInCurScope(curtoken))
		{
			constrainer->reportError(ERROR_SYMBOLDEFINED);
		}

		//make sure terminated right
		nexttoken = scanner->getNextToken();
		
		//check to see if array
		if(nexttoken == T_LBRACKET)
		{
			long arraylength;
			nexttoken = scanner->getNextToken();
			if(nexttoken == T_ENDOFFILE)
			{
				constrainer->reportError(ERROR_UNEXPECTEDENDOFFILE);
				return false;
			}
			else if(!(nexttoken & T_IMMEDIATE))
				constrainer->reportError(ERROR_CONSTANTREQUIRED);

			//remove immediate flag
			nexttoken &= (~(TOKEN_ID)T_IMMEDIATE);

			//check for type
			if(nexttoken != T_CHAR && nexttoken != T_SHORT && nexttoken != T_LONG)
				constrainer->reportError(ERROR_INVALIDTYPE);

			//get length of array
			arraylength = scanner->getImmediateLong();

			//get trailing bracket
			nexttoken = scanner->getNextToken();
			//make sure not error
			if(nexttoken == T_ENDOFFILE)
			{
				constrainer->reportError(ERROR_UNEXPECTEDENDOFFILE);
				return false;
			}
			if(nexttoken != T_RBRACKET)
				constrainer->reportError(ERROR_BRACKETREQUIRED);

			//add array to symbol table
			long defnum = symboltable->getNextDefNum(curtoken);
			vardefstack->pushVarDef(curtoken, defnum);
			symboltable->addSymbol(curtoken, defnum, scanner->getEndNameofLastGet(),
				(arraylength != 0) ? (A_ARRAY | A_USERTYPE) : (A_ARRAY | A_RESIZEABLE),
				curclass, 0, (unsigned char *)"\0\0\0\0\0\0\0\0", curparent);

			//set size of class
			symboltable->setSize(curtoken, defnum, (arraylength != 0) ? (arraylength * symboltable->getSize(curclass, curclassdefnum)) : SIZE_LONG);


			//add to class if that's the case
			if(inclass)
			{
				//keep track of position within class
				symboltable->setAddress(curtoken, defnum, symboltable->getSize(curtoken, defnum));
				//add size to class
				symboltable->addSize(curparent, curparentdefnum, symboltable->getSize(curtoken, defnum));
			}
			else //global
			{
				//keep track of position relative to globals
				symboltable->setAddress(curtoken, defnum, globalvarsize);
				//add size to global size
				globalvarsize += symboltable->getSize(curtoken, defnum);
			}

			nexttoken = scanner->getNextToken();
		}
		else //non-array declaration
		{
			//add variable to symbol table
			long defnum = symboltable->getNextDefNum(curtoken);
			vardefstack->pushVarDef(curtoken, defnum);
			symboltable->addSymbol(curtoken, defnum, scanner->getEndNameofLastGet(),
				A_USERTYPE, curclass, 0, (unsigned char *)"\0\0\0\0\0\0\0\0", curparent);

			//set size of class
			symboltable->setSize(curtoken, defnum, symboltable->getSize(curclass, curclassdefnum));

			//add to class if that's the case
			if(inclass)
			{
				//keep track of position within class
				symboltable->setAddress(curtoken, defnum, symboltable->getSize(curparent, curparentdefnum));
				//add size to class
				symboltable->addSize(curparent, curparentdefnum, symboltable->getSize(curtoken, defnum));
			}
			else //global
			{
				//keep track of position relative to globals
				symboltable->setAddress(curtoken, defnum, globalvarsize);
				//add size to global size
				globalvarsize += symboltable->getSize(curtoken, defnum);
			}
		}

		if(nexttoken == T_ENDOFFILE)
		{
			constrainer->reportError(ERROR_UNEXPECTEDENDOFFILE);
			return false;
		}
		else if(nexttoken != T_SEMICOLON)
			constrainer->reportError(ERROR_INVALIDSYMBOL);

	}
	else
		constrainer->reportError(ERROR_TYPENOTSPECIFIED);

	//return to default
	curattrib = A_SIGNED;

	return true;
}

