#include "common.h"
#include <stdio.h>
#include <math.h>

Scanner::Scanner(char *file, StringTable *st)
{
	stringtable = st;
	strcpy(filename, file);
	codepos = 0;				//start at beginning of file
	curlinenumber = 1;
	includedepth = 0;		//assume at root of include (changed if not)
	included_file = NULL;	//start in current file

	//load source from file
	FILE *f;
	f = fopen(filename, "rb");
	fseek(f, 0, SEEK_END);
	codesize = ftell(f);
	fseek(f, 0, SEEK_SET);
	code = new unsigned char[codesize];
	fread(code, 1, codesize, f);
	fclose(f);
}

Scanner::~Scanner()
{
	delete [] code;
}

TOKEN_ID Scanner::getNextToken()
{
	//if in included file
	if(included_file != NULL)
	{
		TOKEN_ID retval = included_file->getNextToken();
		if(retval != T_ENDOFFILE)
			return retval;
		else //delete included file and continue parsing this one
		{
			delete included_file;
			included_file = NULL;
		}
	}

	//cut out garbage before next token (comments and whitespace)
	for(;;)
	{
		//skip over leading whitespace
		while(codepos < codesize &&
				(code[codepos] == ' ' || code[codepos] == CR || code[codepos] == LF || code[codepos] == TAB) )
		{
			if(code[codepos] == CR)
				curlinenumber++;
			codepos++;
		}

		//comment must be at least 2 characters long
		if(codepos + 2 > codesize)
			break;

		//check for comments before next token
		if(code[codepos] == '/')
		{
			if(code[codepos+1] == '/')
			{
				//get line comment until carriage return
				while(codepos < codesize && code[codepos] != CR)
					codepos++;
				continue;
			}
			else if(code[codepos+1] == '*')
			{
				//get bounded comment
				while(codepos + 1 < codesize && !(code[codepos] == '*' && code[codepos+1] == '/') )
				{
					codepos++;
					if(code[codepos] == CR)
						curlinenumber++;
				}
				codepos += 2;
				continue;
			}
		}

		//done with garbage
		break;
	} //eat garbage

	//mark beginning of current token incase needed for error
	prevtokenpos = codepos;

	//check end of file
	if(codepos >= codesize)
		return T_ENDOFFILE;


	//next character

	//see if identifier
	if( (code[codepos] >= 'a' && code[codepos] <= 'z')
		|| (code[codepos] >= 'A' && code[codepos] <= 'Z')
		|| code[codepos] == '_')
	{
		//process identifier
		char buffer[256];
		int pos = 0;

		//while still in string, copy to buffer
		while(codepos < codesize &&
			( (code[codepos] >= 'a' && code[codepos] <= 'z')
			|| (code[codepos] >= 'A' && code[codepos] <= 'Z')
			|| (code[codepos] >= '0' && code[codepos] <= '9')
			|| code[codepos] == '_') )
		{
			buffer[pos++] = code[codepos++];
		}
		buffer[pos] = '\0';

		TOKEN_ID retval = stringtable->getId(buffer);
		curStringEndName = stringtable->getEndNameofLastGet();
		return retval;
	}

	//check to see if number
	else if( (code[codepos] >= '0' && code[codepos] <= '9')
		|| (codepos+1 < codesize && code[codepos] == '.'	//check to see if float
			&& (code[codepos+1] >= '0' && code[codepos+1] <= '9') ) )
	{
		//process number

		//check if hex (make sure not float, and make sure has preceding '0x')
		if((code[codepos] != '.') && (codepos + 1 < codesize) && code[codepos+1] == 'x')
		{
			long num = 0;
			//hex
			codepos += 2; //skip over hex intro
			//while still in number, accumulate
			while(codepos < codesize
				&& ( (code[codepos] >= '0' && code[codepos] <= '9')
					|| (code[codepos] >= 'A' && code[codepos] <= 'F') ) )
			{
				num <<= 4;
				if(code[codepos] >= '0' && code[codepos] <= '9')
				{
					num += (code[codepos++] - '0');
				}
				else
				{
					num += (code[codepos++] - 'A' + 0x0A);
				}
			} //while parsing hex number

			curImmediateLong = num;
			return T_IMMEDIATE | T_LONG;
		}
		else //not hex
		{
			long num = 0;
			double fnum;
			bool isfloat = false;
			//while still in number, accumulate
			while(codepos < codesize
				&& (code[codepos] >= '0' && code[codepos] <= '9') )
			{
				num *= 10;
				num += (code[codepos++] - '0');
			}

			//check if float
			if(codepos < codesize && code[codepos] == '.')
			{
				codepos++;
				//convert number to float
				fnum = (double)num;
				isfloat = true;

				int divvalue = 1;
				//read in decimal portion
				while(codepos < codesize
					&& (code[codepos] >= '0' && code[codepos] <= '9') )
				{
					divvalue *= 10;
					fnum += ((double)(code[codepos++] - '0')) / divvalue;
				}
			}

			//check if exponentiated float (xx.xxExx)
			if(codepos < codesize && (code[codepos] == 'e' || code[codepos] == 'E') )
			{
				codepos++;
				//convert number to float
				if(!isfloat)
				{
					fnum = (double)num;
					isfloat = true;
				}

				int expval = 0;
				bool negexp = false;
				//see if negative
				if(codepos < codesize && code[codepos] == '-')
				{
					negexp = true;
					codepos++;
				}
				else if(codepos < codesize && code[codepos] == '+')
				{
					codepos++;
				}

				//read in decimal portion
				while(codepos < codesize
					&& (code[codepos] >= '0' && code[codepos] <= '9') )
				{
					expval *= 10;
					expval += (code[codepos++] - '0');
				}
				if(negexp)
					expval = -expval;

				fnum *= pow(10, expval);
			}


			//return correct val
			if(!isfloat)
			{
				curImmediateLong = num;
				return T_IMMEDIATE | T_LONG;
			}
			else
			{
				curImmediateDouble = fnum;
				return T_IMMEDIATE | T_DOUBLE;
			}
		} //not hex
	} //if number
	//symbol
	else
	{
		switch(code[codepos])
		{
			//include all symbols, such as @, *, and string and character literal
			//indicators such as ' and "
		case '.':
			codepos++;
			return T_MEMBEROF;
		case ':':
			codepos++;
			if(codepos < codesize && code[codepos] == ':')
			{
				codepos++;
				return T_SCOPE_RES;
			}
			else if(codepos < codesize && code[codepos] == '=')
			{
				codepos++;
				return T_ARRAYRESIZE;
			}
			else
			{
				return T_INVALID;
			}

		case '+':
			codepos++;
			return T_PLUS;
		case '-':
			codepos++;
			return T_MINUS;
		case '*':
			codepos++;
			return T_MULTIPLY;
		case '/':
			codepos++;
			return T_DIVIDE;
		case '%':
			codepos++;
			return T_MODULUS;

		case '^':
			codepos++;
			return T_BXOR;
		case '~':
			codepos++;
			return T_BNOT;

		case '=':
			codepos++;
			if(codepos < codesize && code[codepos] == '=')
			{
				codepos++;
				return T_EQUALS;
			}
			else
				return T_GETS;
		case '<':
			codepos++;
			if(codepos < codesize && code[codepos] == '=')
			{
				codepos++;
				return T_LEQUAL;
			}
			else if(codepos < codesize && code[codepos] == '<')
			{
				codepos++;
				return T_SHIFTLEFT;
			}
			else
				return T_LESS;
		case '>':
			codepos++;
			if(codepos < codesize && code[codepos] == '=')
			{
				codepos++;
				return T_GEQUAL;
			}
			else if(codepos < codesize && code[codepos] == '>')
			{
				codepos++;
				return T_SHIFTRIGHT;
			}
			else
				return T_GREATER;
		case '!':
			codepos++;
			if(codepos < codesize && code[codepos] == '=')
			{
				codepos++;
				return T_NOTEQUAL;
			}
			else
				return T_ENOT;

		case '|':
			codepos++;
			if(codepos < codesize && code[codepos] == '|')
			{
				codepos++;
				return T_EOR;
			}
			else
				return T_BOR;
		case '&':
			codepos++;
			if(codepos < codesize && code[codepos] == '&')
			{
				codepos++;
				return T_EAND;
			}
			else
				return T_BAND;
		case '(':
			codepos++;
			return T_LPAREN;
		case ')':
			codepos++;
			return T_RPAREN;

		case ';':
			codepos++;
			return T_SEMICOLON;
		case ',':
			codepos++;
			return T_COMMA;

		case '{':
			codepos++;
			return T_LBRACE;
		case '}':
			codepos++;
			return T_RBRACE;

		case '@':
			codepos++;
			return T_PARALLEL;

		case '[':
			codepos++;
			return T_LBRACKET;
		case ']':
			codepos++;
			return T_RBRACKET;

		case '\'':	//immediate character
			codepos++;
			{
				if(codepos < codesize)
				{
					if(code[codepos] != '\\')
					{
						curImmediateLong = code[codepos];
						codepos++;
						//eliminate extra trailer apostrophe if there
						if(code[codepos] == '\'')
							codepos++;
						return T_IMMEDIATE | T_CHAR;
					} //regular character
					else
					{
						codepos++;
						if(codepos < codesize)
						{
							switch(code[codepos])
							{
							case '0':
								curImmediateLong = (long)('\0');
								break;
							case 'n':
								curImmediateLong = (long)CR;
								break;
							case 't':
								curImmediateLong = (long)TAB;
								break;
							case '\'':
								curImmediateLong = (long)('\'');
								break;
							case '\\':
								curImmediateLong = (long)('\\');
								break;
							default:	//unrecognized character
								return T_INVALID;
							}

							codepos++;
							//eliminate extra trailer apostrophe if there
							if(code[codepos] == '\'')
								codepos++;
							return T_IMMEDIATE | T_CHAR;
						} //still in file
						else
						{
							return T_INVALID;
						}
					} //escape character
				} //still in file
				else
				{
					return T_INVALID;	//return invalid if apostrophe at end of file
				}
			}
		case '"':	//immediate string
			codepos++;
			{
				int tempcodepos = codepos;
				int slen = 0;
				//count length of string -ok if it's a little longer (with escape chars)
				while(tempcodepos < codesize && code[tempcodepos] != '"')
				{
					if(code[tempcodepos] == '\\')
						tempcodepos++;

					slen++;
					tempcodepos++;
				}

				if(tempcodepos == codesize)
				{
					codepos++;
					return T_INVALID;
				}

				//now create string
				curImmediateStr = new char[slen+1];
				slen = 0;

				//assured that is within file
				while(/*codepos < codesize && */code[codepos] != '"')
				{
					if(code[codepos] != '\\')
					{
						curImmediateStr[slen] = code[codepos];
					}
					else
					{
						codepos++;
						switch(code[codepos])
						{
						case '0':
							curImmediateStr[slen] = (long)('\0');
							break;
						case 'n':
							curImmediateStr[slen] = (long)CR;
							break;
						case 't':
							curImmediateStr[slen] = (long)TAB;
							break;
						case '"':
							curImmediateStr[slen] = (long)('"');
							break;
						case '\\':
							curImmediateStr[slen] = (long)('\\');
							break;
						default:	//unrecognized character
							return T_INVALID;
						}
					}					

					slen++;
					codepos++;
				}

				//skip trailing "
				codepos++;
				//terminate string
				curImmediateStr[slen] = '\0';
				return T_IMMEDIATE | T_IMMEDIATE_STR;
			}

		case '#':
			//preprocessor directive
			{
				//#include preprocessor directive
				codepos++;
				if(codepos+9 > codesize)
				{
					//don't have enough to include
					return T_INVALID;
				}

				char buffer[256];
				int buflen = 0;
				//go to start of filename
				codepos += 8;
				while(codepos < codesize && code[codepos] != CR)
				{
					buffer[buflen++] = code[codepos++];
				}
				buffer[buflen] = '\0';

				//curlinenumber++; <-taken care of when exit include file

				included_file = new Scanner(buffer, stringtable);
				//check for include recursion (bad)
				included_file->includedepth = includedepth + 1;
				if(included_file->includedepth > MAX_INCLUDE_RECURSION)
				{
					delete included_file;
					included_file = NULL;
					return T_RECURSEINCLUDE;
				}


				//now get next token from new file
				TOKEN_ID retval = included_file->getNextToken();
				if(retval != T_ENDOFFILE)
					return retval;
				else //empty file -delete and continue parsing this one
				{
					delete included_file;
					included_file = NULL;
					//need to get next token... yucky recursion, but necessary without goto
					return getNextToken();
				}
			}
			return T_INVALID;

		default:
			codepos++;
			return T_INVALID;
		}
	} //symbol

	codepos++;
	return T_INVALID;
}


StringTrieNode *Scanner::getEndNameofLastGet()
{
	if(included_file != NULL)
	{
		return included_file->getEndNameofLastGet();
	}
	else
	{
		return curStringEndName;
	}
}

//if getNextToken returns with a T_IMMEDIATE flag, call this
char *Scanner::getImmediateString()
{
	if(included_file != NULL)
	{
		return included_file->getImmediateString();
	}
	else
	{
		return curImmediateStr;
	}
}

long Scanner::getImmediateLong()
{
	if(included_file != NULL)
	{
		return included_file->getImmediateLong();
	}
	else
	{
		return curImmediateLong;
	}
}

double Scanner::getImmediateDouble()
{
	if(included_file != NULL)
	{
		return included_file->getImmediateDouble();
	}
	else
	{
		return curImmediateDouble;
	}
}

int Scanner::getCurPos()
{
	if(included_file != NULL)
	{
		return included_file->getCurPos();
	}
	else
	{
		return codepos;
	}
}

char *Scanner::SourceFile()
{
	if(included_file != NULL)
	{
		return included_file->SourceFile();
	}
	else
	{
		return filename;
	}
}

//used for error information
int Scanner::getCurrentLineNumber()
{
	if(included_file != NULL)
	{
		return included_file->getCurrentLineNumber();
	}
	else
	{
		return curlinenumber;
	}
}

char *Scanner::getCurrentLineWithPosition()
{
	if(included_file != NULL)
	{
		return included_file->getCurrentLineWithPosition();
	}
	else
	{
		static char buffer[512];	//static so it stays there -can access outside
		//get before current
		int temppos;
		for(temppos = prevtokenpos; temppos > 0 && (code[temppos] != CR && code[temppos] != LF); temppos--);
		if(code[temppos] == CR || code[temppos] == LF)
			temppos++;	//don't write out if there
		memset(buffer, 0, 512);
		memcpy(buffer, &code[temppos], prevtokenpos - temppos);

		strcat(buffer, " <*> ");

		//copy after current
		int bufpos;
		for(temppos = prevtokenpos, bufpos = strlen(buffer); bufpos < 512 && temppos < codesize && code[temppos] != CR; bufpos++, temppos++)
			buffer[bufpos] = code[temppos];
		return buffer;
	}
}

