#include "common.h"

/////////////////////////////////////////////////////////////
//String Table

StringTable::StringTable()
{
	//cannot have string length of zero
	//so set to not a token
	root = new StringTrieNode;
	root->id = 0;
	root->parent = NULL;
	int i;
	for(i = 0; i < ST_NUM_ACCEPT_CHARS; i++)
		root->list[i] = NULL;

	//first block of tokens are reserved
	nextId = T_IDENTIFIERSTART;

	//add reserved words into table
	for(i = 0; i < numReservedWords; i++)
	{
		addReservedWord(reservedWords[i].rw, reservedWords[i].id);
	}

}

TOKEN_ID StringTable::getId(const char *str)
{
	lastentry = root;
	for(int i = 0; str[i] != '\0'; i++)
	{
		int p = letterPosition(str[i]);

		//create new portion of the tree if need to
		if(lastentry->list[p] == NULL)
		{
			StringTrieNode *tempnode = new StringTrieNode;
			tempnode->id = 0;
			tempnode->parent = lastentry;
			for(int j = 0; j < ST_NUM_ACCEPT_CHARS; j++)
				tempnode->list[j] = NULL;

			lastentry->list[p] = tempnode;
		}

		//traverse tree
		lastentry = lastentry->list[p];
	}

	//add to list if not already
	if(lastentry->id == 0)
		lastentry->id = nextId++;

	//return token
	return lastentry->id;
}

StringTrieNode *StringTable::getEndNameofLastGet()
{
	return lastentry;
}

char *StringTable::getString(StringTrieNode *endname)
{
	char buffer[128];
	//start from the end and work towards the front
	buffer[127] = '\0';
	int curpos = 127;

	const char lettertable[] = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_";

	//go upwards through tree to accumulating name
	while(endname->parent != NULL)
	{
		for(int i = 0; i < ST_NUM_ACCEPT_CHARS; i++)
		{
			if(endname->parent->list[i] != endname)
				continue;
			else //found it
			{
				buffer[--curpos] = lettertable[i];
				break;
			}
		}		

		endname = endname->parent;
	}

	//copy name from buffer
	char *ret = new char[128 - curpos];
	strcpy(ret, &buffer[curpos]);
	return ret;
}

void StringTable::addReservedWord(const char *str, TOKEN_ID id)
{
	lastentry = root;
	for(int i = 0; str[i] != '\0'; i++)
	{
		int p = letterPosition(str[i]);

		//create new portion of the tree if need to
		if(lastentry->list[p] == NULL)
		{
			StringTrieNode *tempnode = new StringTrieNode;
			tempnode->id = 0;
			tempnode->parent = lastentry;
			for(int j = 0; j < ST_NUM_ACCEPT_CHARS; j++)
				tempnode->list[j] = NULL;

			lastentry->list[p] = tempnode;
		}

		//traverse tree
		lastentry = lastentry->list[p];
	}
	lastentry->id = id;
}

int StringTable::letterPosition(const char c)
{
	//order:
	//abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_
	if( c >= 'a' && c <= 'z' )
		return (c - 'a');

	if( (c >= 'A') && (c <= 'Z') )
		return (c - 'A' + 26);

	if( (c >= '0') && (c <= '9') )
		return (c - '0' + 52);

	if(c == '_')
		return 62;

	return -1;
}


StringTable::~StringTable()
{
	DeleteNode(root);
}

//deletes node and children
void StringTable::DeleteNode(StringTrieNode *n)
{
	for(int i = 0; i < ST_NUM_ACCEPT_CHARS; i++)
	{
		if(n->list[i] == NULL)
			continue;
		DeleteNode(n->list[i]);
	}
	delete n;
}


/////////////////////////////////////////////////////////////
//Symbol Table

SymbolTable::SymbolTable()
{
	for(int i = 0; i < SYMBOL_TABLE_HASH_SIZE; i++)
		hashTable[i] = NULL;

	//set these to first, just incase called on accident
	CurFuncHashPos = 0;
	CurFuncNode = NULL;
	CurGlobalHashPos = 0;
	CurGlobalNode = NULL;
}

void SymbolTable::DeleteNode(ASTNode *n)
{
	if(n == NULL)
		return;

//***this causes a warning...
//	if(n->op & T_IMMEDIATE)
//		delete n->data;

/*
	if(n->op == T_FUNCTIONCALL || n->op == T_PFUNCTIONCALL)
	{
		for(int i = 0; i < n->funccall->numparameters; i++)
			DeleteNode(n->funccall->parameters[i]);
		delete n->funccall;
	}

	for(int i = 0; i < 3; i++)
	{
		if(n->child[i] != NULL)
			DeleteNode(n->child[i]);
	}
	delete n;
*/
}

SymbolTable::~SymbolTable()
{
	//remove each entry from symbol table
	for(int i = 0; i < SYMBOL_TABLE_HASH_SIZE; i++)
	{
		SymbolTableNode *temp;
		temp = hashTable[i];
		while(temp != NULL)
		{
			hashTable[i] = temp->next;
			if(temp->functype != NULL)
			{
				//delete representations of function

				//delete AST code
				DeleteNode(temp->functype->function);

				//delete MIR code
				MIRInstr *mtemp;
				while(temp->functype->mircode != NULL)
				{
					mtemp = temp->functype->mircode->next;
					delete temp->functype->mircode;
					temp->functype->mircode = mtemp;
				}

				delete temp->functype;
			}
			delete temp;
			temp = hashTable[i];
		}
	}
}

void SymbolTable::addSymbol(TOKEN_ID id, long defnum, StringTrieNode *nameend, ATTRIB_TYPE type, TOKEN_ID usertype, ATTRIB_ADDR address, unsigned char *initval, TOKEN_ID parent)
{
	int pos = (id & SYMBOL_TABLE_HASH_AND);
	SymbolTableNode *n = new SymbolTableNode;
	n->id = id;
	n->defnum = defnum;
	n->nameend = nameend;
	n->type = type;
	n->usertype = usertype;
	n->address = address;
	n->parent = parent;
	n->functype = NULL;
	switch(type & 0xFF)		//initialize and set to size of type (not an array)
	{
	case T_CHAR:
		n->initval[0] = initval[0];
		n->size = SIZE_CHAR;
		break;
	case T_SHORT:
		n->initval[0] = initval[0];
		n->initval[1] = initval[1];
		n->size = SIZE_SHORT;
		break;
	case T_LONG:
		n->initval[0] = initval[0];
		n->initval[1] = initval[1];
		n->initval[2] = initval[2];
		n->initval[3] = initval[3];
		n->size = SIZE_LONG;
		break;
	case T_FLOAT:
		n->initval[0] = initval[0];
		n->initval[1] = initval[1];
		n->initval[2] = initval[2];
		n->initval[3] = initval[3];
		n->size = SIZE_FLOAT;
		break;
	case T_DOUBLE:
		n->initval[0] = initval[0];
		n->initval[1] = initval[1];
		n->initval[2] = initval[2];
		n->initval[3] = initval[3];
		n->initval[4] = initval[4];
		n->initval[5] = initval[5];
		n->initval[6] = initval[6];
		n->initval[7] = initval[7];
		n->size = SIZE_DOUBLE;
		break;
	case T_CLASS:
	case T_PCLASS:
		n->size = 0;	//set to zero for time being -will be set later
		break;
	default:
		n->size = SIZE_LONG;	//assume pointer
		break;
	}
	//must be set from outside symbol table
	n->functype = NULL;

	n->next = hashTable[pos];
	hashTable[pos] = n;
}

bool SymbolTable::isSymbolDefined(TOKEN_ID id, long defnum)
{
	//returns NULL (false) if not defined, nonzero if defined
	return (findNode(id, defnum) != 0);
}

//get next usable defnum (for newer identifier)
long SymbolTable::getNextDefNum(TOKEN_ID id)
{
	long max = 1;
	for(SymbolTableNode *n = hashTable[id & SYMBOL_TABLE_HASH_AND]; n != NULL; n = n->next)
	{
		if(n->id == id && n->defnum >= max)
			max = n->defnum + 1;
	}
	return max;
}

//set functions
void SymbolTable::setType(TOKEN_ID id, long defnum, ATTRIB_TYPE type)
{
	SymbolTableNode *n = findNode(id, defnum);
	n->type = type;
}

void SymbolTable::setUserType(TOKEN_ID id, long defnum, TOKEN_ID usertype)
{
	SymbolTableNode *n = findNode(id, defnum);
	n->usertype = usertype;
}

void SymbolTable::setAddress(TOKEN_ID id, long defnum, ATTRIB_ADDR addr)
{
	SymbolTableNode *n = findNode(id, defnum);
	n->address = addr;
}

void SymbolTable::setInitVal(TOKEN_ID id, long defnum, unsigned char *initval, ATTRIB_TYPE type)
{
	SymbolTableNode *n = findNode(id, defnum);

	if(n->type & A_ARRAY)
	{
		//keep pointer to array
		((unsigned char **)n->initval)[0] = initval;
		return;
	}

	//must type cast if types do not match
	switch(n->type & A_BASETYPES)
	{
	case T_CHAR:
		switch(type)
		{
		case T_CHAR:
			((char *)n->initval)[0] = (char)((char *)initval)[0];
			break;
		case T_SHORT:
			((char *)n->initval)[0] = (char)((short *)initval)[0];
			break;
		case T_LONG:
			((char *)n->initval)[0] = (char)((long *)initval)[0];
			break;
		case T_FLOAT:
			((char *)n->initval)[0] = (char)((float *)initval)[0];
			break;
		case T_DOUBLE:
			((char *)n->initval)[0] = (char)((double *)initval)[0];
			break;
		}
		break;
	case T_SHORT:
		switch(type)
		{
		case T_CHAR:
			((short *)n->initval)[0] = (short)((char *)initval)[0];
			break;
		case T_SHORT:
			((short *)n->initval)[0] = (short)((short *)initval)[0];
			break;
		case T_LONG:
			((short *)n->initval)[0] = (short)((long *)initval)[0];
			break;
		case T_FLOAT:
			((short *)n->initval)[0] = (short)((float *)initval)[0];
			break;
		case T_DOUBLE:
			((short *)n->initval)[0] = (short)((double *)initval)[0];
			break;
		}
		break;
	case T_LONG:
		switch(type)
		{
		case T_CHAR:
			((long *)n->initval)[0] = (long)((char *)initval)[0];
			break;
		case T_SHORT:
			((long *)n->initval)[0] = (long)((short *)initval)[0];
			break;
		case T_LONG:
			((long *)n->initval)[0] = (long)((long *)initval)[0];
			break;
		case T_FLOAT:
			((long *)n->initval)[0] = (long)((float *)initval)[0];
			break;
		case T_DOUBLE:
			((long *)n->initval)[0] = (long)((double *)initval)[0];
			break;
		}
		break;
	case T_FLOAT:
		switch(type)
		{
		case T_CHAR:
			((float *)n->initval)[0] = (float)((char *)initval)[0];
			break;
		case T_SHORT:
			((float *)n->initval)[0] = (float)((short *)initval)[0];
			break;
		case T_LONG:
			((float *)n->initval)[0] = (float)((long *)initval)[0];
			break;
		case T_FLOAT:
			((float *)n->initval)[0] = (float)((float *)initval)[0];
			break;
		case T_DOUBLE:
			((float *)n->initval)[0] = (float)((double *)initval)[0];
			break;
		}
		break;
	case T_DOUBLE:
		switch(type)
		{
		case T_CHAR:
			((double *)n->initval)[0] = (double)((char *)initval)[0];
			break;
		case T_SHORT:
			((double *)n->initval)[0] = (double)((short *)initval)[0];
			break;
		case T_LONG:
			((double *)n->initval)[0] = (double)((long *)initval)[0];
			break;
		case T_FLOAT:
			((double *)n->initval)[0] = (double)((float *)initval)[0];
			break;
		case T_DOUBLE:
			((double *)n->initval)[0] = (double)((double *)initval)[0];
			break;
		}
		break;
	default:
		break;
	}
}

void SymbolTable::setParent(TOKEN_ID id, long defnum, TOKEN_ID parent)
{
	SymbolTableNode *n = findNode(id, defnum);
	n->parent = parent;
}

void SymbolTable::setFunctionType(TOKEN_ID id, long defnum, FunctionType *f)
{
	SymbolTableNode *n = findNode(id, defnum);
	n->functype = f;
}

void SymbolTable::setSize(TOKEN_ID id, long defnum, long size)	//useful for arrays
{
	SymbolTableNode *n = findNode(id, defnum);
	n->size = size;
}

void SymbolTable::addSize(TOKEN_ID id, long defnum, long size)	//useful for classes
{
	SymbolTableNode *n = findNode(id, defnum);
	n->size += size;
}

//get functions
ATTRIB_TYPE SymbolTable::getType(TOKEN_ID id, long defnum)
{
	SymbolTableNode *n = findNode(id, defnum);
	if(n == NULL)
		return 0;
	return n->type;
}

TOKEN_ID SymbolTable::getUserType(TOKEN_ID id, long defnum)
{
	SymbolTableNode *n = findNode(id, defnum);
	if(n == NULL)
		return 0;
	return n->usertype;
}

ATTRIB_ADDR SymbolTable::getAddress(TOKEN_ID id, long defnum)
{
	SymbolTableNode *n = findNode(id, defnum);
	if(n == NULL)
		return 0;
	return n->address;
}

unsigned char *SymbolTable::getInitVal(TOKEN_ID id, long defnum)
{
	SymbolTableNode *n = findNode(id, defnum);
	if(n == NULL)
		return NULL;
	return n->initval;
}

StringTrieNode *SymbolTable::getNameLookup(TOKEN_ID id, long defnum)
{
	SymbolTableNode *n = findNode(id, defnum);
	if(n == NULL)
		return NULL;
	return n->nameend;
}

TOKEN_ID SymbolTable::getParent(TOKEN_ID id, long defnum)
{
	SymbolTableNode *n = findNode(id, defnum);
	if(n == NULL)
		return 0;
	return n->parent;
}

FunctionType *SymbolTable::getFunctionType(TOKEN_ID id, long defnum)
{
	SymbolTableNode *n = findNode(id, defnum);
	if(n == NULL)
		return NULL;
	return n->functype;
}

long SymbolTable::getSize(TOKEN_ID id, long defnum)
{
	SymbolTableNode *n = findNode(id, defnum);
	if(n == NULL)
		return 0;
	return n->size;
}

//useful for getting IDs of functions of a class or global
long SymbolTable::getDefnum(TOKEN_ID id, TOKEN_ID parent)
{
	for(SymbolTableNode *n = hashTable[id & SYMBOL_TABLE_HASH_AND]; n != NULL; n = n->next)
	{
		if(n->id == id && n->parent == parent)
		{
			return n->defnum;
		}
	}
	return INVALIDDEFNUM;
}

//returns defnum, otherwise 0 if none
long SymbolTable::getMatchingFuncDefnum(TOKEN_ID id, FunctionType* f, TOKEN_ID parent)
{
	for(SymbolTableNode *n = hashTable[id & SYMBOL_TABLE_HASH_AND]; n != NULL; n = n->next)
	{
		if(n->id == id && n->parent == parent)
		{
			//check function types
			if(n->functype != NULL && n->functype->numarguements == f->numarguements)
			{
				int i;
				for(i = 0; i < f->numarguements; i++)
				{
					if(f->arguements[i] != n->functype->arguements[i])
						break;
				}
				if(i == f->numarguements)
					return n->defnum;
			}//check function
		}//check id
	}
	return INVALIDDEFNUM;
}

//returns defnum, otherwise 0 if none
long SymbolTable::getClosestFuncDefnum(TOKEN_ID id, FunctionType* f, TOKEN_ID parent)
{
	long closestfuncdefnum = INVALIDDEFNUM;
	int closest_numexact = 0;

	for(SymbolTableNode *n = hashTable[id & SYMBOL_TABLE_HASH_AND]; n != NULL; n = n->next)
	{
		if(n->id == id && n->parent == parent)
		{
			//check function types
			if(n->functype != NULL && n->functype->numarguements == f->numarguements)
			{
				int i;
				//make sure no arguements are of a more general type than parameters
				for(i = 0; i < f->numarguements; i++)
				{
					if(n->functype->arguements[i]
							!= getGeneralType(n->functype->arguements[i], f->arguements[i]) )
						break;
				}
				if(i != f->numarguements)
					continue;

				//find number of exact arguement/parameter matches
				for(i = 0; i < f->numarguements; i++)
				{
					if(f->arguements[i] != n->functype->arguements[i])
						break;
				}	
				if(i >= closest_numexact)
					closestfuncdefnum = n->defnum;

			}//check function
		}//check id
	}
	return closestfuncdefnum;
}

//gets defnum for class with respective id
long SymbolTable::getClassDefnum(TOKEN_ID id)
{
	for(SymbolTableNode *n = hashTable[id & SYMBOL_TABLE_HASH_AND]; n != NULL; n = n->next)
	{
		if(n->id == id && (n->type == T_CLASS || n->type == T_PCLASS))
		{
			return n->defnum;
		}
	}
	return INVALIDDEFNUM;
}

//gets defnum of the variable in the class
long SymbolTable::getDefnumOfMemberVariable(TOKEN_ID id, TOKEN_ID class_id)
{
	for(SymbolTableNode *n = hashTable[id & SYMBOL_TABLE_HASH_AND]; n != NULL; n = n->next)
	{
		if(n->id == id && (n->parent == class_id))
		{
			return n->defnum;
		}
	}
	return INVALIDDEFNUM;
}

//for fast access
SymbolTableNode *SymbolTable::getSymbolTableEntry(TOKEN_ID id, long defnum)
{
	return findNode(id, defnum);
}

//returns pointer to node containing id
SymbolTableNode *SymbolTable::findNode(TOKEN_ID id, long defnum)
{
	for(SymbolTableNode *n = hashTable[id & SYMBOL_TABLE_HASH_AND]; n != NULL; n = n->next)
	{
		if(n->id == id && n->defnum == defnum)
			return n;
	}
	return NULL;
}


//gets first function listed in hash table
FunctionType *SymbolTable::getFirstFunction()
{
	for(CurFuncHashPos = 0; CurFuncHashPos < SYMBOL_TABLE_HASH_SIZE; CurFuncHashPos++)
	{
		for(CurFuncNode = hashTable[CurFuncHashPos]; CurFuncNode != NULL; CurFuncNode = CurFuncNode->next)
		{
			if(CurFuncNode->functype != NULL)
				return CurFuncNode->functype;
		}
	}

	return NULL;	
}

//gets next function listed in hash table
FunctionType *SymbolTable::getNextFunction()
{
	CurFuncNode = CurFuncNode->next; //already did current one
	while(CurFuncHashPos < SYMBOL_TABLE_HASH_SIZE)
	{
		while(CurFuncNode != NULL)
		{
			if(CurFuncNode->functype != NULL)
				return CurFuncNode->functype;
			CurFuncNode = CurFuncNode->next;
		}
		CurFuncHashPos++;
		CurFuncNode = hashTable[CurFuncHashPos];
	}
	return NULL;
}

TOKEN_ID SymbolTable::getCurFunctionID()
{
	return CurFuncNode->id;
}

long SymbolTable::getCurFunctionDefnum()
{
	return CurFuncNode->defnum;
}


void SymbolTable::resetGlobalSelection()
{
	CurGlobalHashPos = 0;
	CurGlobalNode = hashTable[CurGlobalHashPos];
}

bool SymbolTable::selectNextGlobal()
{
	if(CurGlobalNode != NULL)
		CurGlobalNode = CurGlobalNode->next; //already did current one
	while(CurGlobalHashPos < SYMBOL_TABLE_HASH_SIZE)
	{
		while(CurGlobalNode != NULL)
		{
			if(CurGlobalNode->parent == GLOBALSCOPE && CurGlobalNode->functype == NULL)
				return true;
			CurGlobalNode = CurGlobalNode->next;
		}
		CurGlobalHashPos++;
		CurGlobalNode = hashTable[CurGlobalHashPos];
	}
	return false;
}

TOKEN_ID SymbolTable::getCurGlobalID()
{
	return CurGlobalNode->id;
}

long SymbolTable::getCurGlobalDefnum()
{
	return CurGlobalNode->defnum;
}

